import numpy as _np
import tensorflow as _tf
import pandas as _pd
import so_ml_tools as _soml

from typing import Union as _Union


def is_binary_classification(y: _Union[list, _np.ndarray, _pd.Series, _pd.DataFrame, _tf.Tensor]) -> bool:
    """
        Checks if the input array represents a multiclass classification problem.

        Parameters:
        y : Union[list, numpy.ndarray, pandas.Series, pandas.DataFrame, tensorflow.Tensor]
            The input array representing the target variable.

        Returns:
        bool
            True if the input array represents a multiclass classification problem, False otherwise.
        """
    if not isinstance(y, _np.ndarray):
        y = _soml.util.types.to_numpy(value=y)

    if y.size == 0:
        return False

    # When there is one dimension we can have two situations, one is they are probabilities so all
    # values have a real number between 0 and (including) 1, or it can be 0 or 1 when we talk about
    # predictions.
    if len(y.shape) == 1:
        return _np.all(_np.logical_and(y >= 0, y <= 1))

    # # When there are two dimensions where the second dimension contains multiple values
    # # we expect a probability type of data, so values should be between 0 and (including) 1
    # if len(y.shape) == 2 and len(y[0]) > 1:
    #     return _np.all(_np.logical_and(y >= 0, y <= 1))

    # When there are two dimensions where the second dimension only contains a single
    # value we expect this to be whole numbers only.
    if len(y.shape) == 2 and len(y[0]) == 1:
        return _np.all(_np.logical_and(y >= 0, y <= 1))

    return False


def is_binary_probability(y: _Union[list, _np.ndarray, _pd.Series, _pd.DataFrame, _tf.Tensor]) -> bool:
    """Check if the input array represents binary probabilities.

        Args:
            y (list or numpy.ndarray or pandas.Series or pandas.DataFrame or tensorflow.Tensor):
                Input array to be checked.

        Returns:
            bool: True if the input array represents binary probabilities, False otherwise.

        Raises:
            None
        """
    if not isinstance(y, _np.ndarray):
        y = _soml.util.types.to_numpy(value=y)

    if not is_binary_classification(y=y):
        return False

    if _np.all((y >= 0) & (y <= 1)):
        return True

    elif len(y.shape) == 2:
        # Check if the second dimension only contains a 1-dimensional array
        if all(isinstance(y_prob, _np.ndarray) and len(y_prob.shape) == 1 for y_prob in y):
            return all(_np.all((y_prob >= 0) & (y_prob <= 1)) for y_prob in y)

    return False


def binary_probability_to_prediction(y: _Union[list, _np.ndarray, _pd.Series, _pd.DataFrame, _tf.Tensor],
                                     maintain_shape: bool = True, threshold: float = 0.5) -> _np.ndarray:
    """Converts binary probabilities to binary predictions based on a threshold.

        Args:
            y (list or numpy.ndarray or pandas.Series or pandas.DataFrame or tensorflow.Tensor):
                Input array of binary probabilities to be converted.
            maintain_shape (bool, optional): Whether to maintain the shape of the input array. Defaults to True.
            threshold (float, optional): Threshold value for binary classification. Defaults to 0.5.

        Returns:
            numpy.ndarray: Array of binary predictions.

        Raises:
            ValueError: If the input array does not represent binary probabilities.
        """
    if not isinstance(y, _np.ndarray):
        y = _soml.util.types.to_numpy(value=y)

    if not is_binary_probability(y):
        raise ValueError('Input must be a binary probability array.')

    y_preds = _np.where(y > threshold, 1, 0).astype(_np.int64)
    if maintain_shape:
        return y_preds

    return y_preds.ravel()


def is_multiclass_classification(y: _Union[list, _np.ndarray, _pd.Series, _pd.DataFrame, _tf.Tensor]) -> bool:
    """Check if the input array represents multiclass classification.

        Args:
            y (list or numpy.ndarray or pandas.Series or pandas.DataFrame or tensorflow.Tensor):
                Input array to be checked.

        Returns:
            bool: True if the input array represents multiclass classification, False otherwise.

        Raises:
            None
        """
    if not isinstance(y, _np.ndarray):
        y = _soml.util.types.to_numpy(value=y)

    if len(y.shape) == 2 and len(y[0]) > 1:
        return _np.all(_np.logical_and(y >= 0, y <= 1))

    if len(y.shape) == 2 and len(y[0]) == 1:
        return _np.all(_np.logical_and(_np.equal(_np.mod(y, 1), 0), y >= 0))

    return False


def is_multiclass_propabilities(y: _Union[list, _np.ndarray, _pd.Series, _pd.DataFrame, _tf.Tensor]) -> bool:
    """Check if the input array represents multiclass probabilities.

        Args:
            y (list or numpy.ndarray or pandas.Series or pandas.DataFrame or tensorflow.Tensor):
                Input array to be checked.

        Returns:
            bool: True if the input array represents multiclass probabilities, False otherwise.

        Raises:
            None
        """
    if not isinstance(y, _np.ndarray):
        y = _soml.util.types.to_numpy(value=y)

    if not is_multiclass_classification(y=y):
        return False

    return _np.all(_np.isclose(_np.sum(y, axis=1), 1.0))


def multiclass_probability_to_prediction(y: _Union[list, _np.ndarray, _pd.Series, _pd.DataFrame, _tf.Tensor],
                                         maintain_shape=False) -> _np.ndarray:
    """Converts multiclass probabilities to multiclass predictions.

    Args:
        y (list or numpy.ndarray or pandas.Series or pandas.DataFrame or tensorflow.Tensor):
            Input array of multiclass probabilities to be converted.
        maintain_shape (bool, optional): Whether to maintain the shape of the input array. Defaults to False.

    Returns:
        numpy.ndarray: Array of multiclass predictions.

    Raises:
        ValueError: If the input array does not represent multiclass probabilities.
    """
    if not isinstance(y, _np.ndarray):
        y = _soml.util.types.to_numpy(value=y)

    if not is_multiclass_propabilities(y):
        raise ValueError('Input must be a multi-class probability array.')

    if maintain_shape:
        return _np.reshape(_np.argmax(y, axis=1), newshape=(y.shape[0], 1))

    return _np.argmax(y, axis=1)


def probability_to_prediction(y: _Union[list, _np.ndarray, _pd.Series, _pd.DataFrame, _tf.Tensor],
                              maintain_shape=None, from_logits=False) -> _np.ndarray:
    """Converts probabilities to binary or multiclass predictions based on the input array shape.

        Args:
            y (list or numpy.ndarray or pandas.Series or pandas.DataFrame or tensorflow.Tensor):
                Input array of probabilities to be converted.
            maintain_shape (bool, optional): Whether to maintain the shape of the input array.
                If None, the function automatically determines the shape based on the input type.
                Defaults to None.

        Returns:
            numpy.ndarray: Array of binary or multiclass predictions.

        Raises:
            ValueError: If the input array does not represent binary or multiclass probabilities.
    """
    if not isinstance(y, _np.ndarray):
        y = _soml.util.types.to_numpy(value=y)

    # In case the values are logits, convert them using softmax.
    if from_logits:
        y = _soml.util.types.to_numpy(value=_tf.nn.softmax(y, axis=1))

    if is_multiclass_classification(y=y):
        if is_multiclass_propabilities(y=y):
            if maintain_shape is None:
                maintain_shape = False
            return multiclass_probability_to_prediction(y, maintain_shape=maintain_shape)
        else:
            return y
    elif is_binary_classification(y=y):
        if is_binary_probability(y=y):
            if maintain_shape is None:
                maintain_shape = True
            return binary_probability_to_prediction(y, maintain_shape=maintain_shape)
        else:
            return y

    raise ValueError('Input must be either a binary or multi-class probability array, is the value a logits value?')



import matplotlib.pyplot as _plt
import seaborn as _sns
import pandas as _pd
import numpy as _np
import scipy as _sp


def count_categories(dataframe: _pd.DataFrame, label_column: str = None, column_names: list[str] = None, cols=3,
                     figsize: tuple = None):
    """
    Create count plots for categorical columns in a DataFrame, optionally segmented by a label column.

    Parameters:
    - dataframe (pd.DataFrame): The input DataFrame.
    - label_column (str): The column used for segmentation (optional). If provided, each count plot will be segmented by the unique values in this column.
    - column_names (List[str]): List of column names to create count plots for. If not provided, all non-numeric columns in the DataFrame will be used.
    - cols (int): Number of columns in the grid layout for count plots.
    - figsize (Tuple[int, int]): Size of the entire figure (width, height). If not provided, the figure size is calculated based on the number of rows and a default height.

    Returns:
    - None

    Example:
    ```python
    import pandas as pd
    import matplotlib.pyplot as plt
    import seaborn as sns

    # Sample DataFrame
    data = {'Category': ['A', 'B', 'A', 'C', 'B', 'C'], 'Label': [1, 0, 1, 1, 0, 1]}
    df = pd.DataFrame(data)

    # Create count plots for categorical columns
    count_categories(dataframe=df, label_column='Label', cols=2)
    ```
    """
    if column_names is None:
        column_names = dataframe.select_dtypes(exclude='number').columns.tolist()

    rows = max(int(len(column_names) / cols), 1)
    cols = min(cols, len(column_names))
    rows += 1 if rows * cols < len(column_names) else 0

    # If figsize is not specified then calculate the fig-size
    if figsize is None:
        figsize = (17, rows * 4)

    fig, axs = _plt.subplots(nrows=rows, ncols=cols, figsize=figsize)

    # If we have more then one column then flatten the axis so we can loop through them,
    # if we have only one column then create list containing the axis so we can still loop through it.
    if len(column_names) > 1:
        axs = axs.flatten()
    else:
        axs = [axs]

    n = 0
    for column_name in column_names:
        _sns.countplot(ax=axs[n], data=dataframe, x=column_name, hue=label_column)
        n += 1

    _plt.show()


def histogram_for_columns(dataframe: _pd.DataFrame, column_names: list[str] = None, log=False,
                          min_nunique: int = 0, max_nunique: int = 50, figsize: tuple = None, cols=1,
                          verbose=1, label_color='black'):
    """
    Plots a histogram for each of the numeric columns in the DataFrame.

    :param dataframe: the pandas dataframe
    :param column_names: columns which exist within the DataFrame if none specified all columns that are numeric
        will be processed
    :param log: set to True to enable logarithmic scaling
    :param min_nunique: minimum number of unique values present, if lower then this then no graph will be displayed (since it is basically boolean)
    :param max_nunique: maximum number of unique values present, only applicable to object column types since these cannot be binned
    :param figsize: size of the plot, if None specified then one is calculated
    :param cols: number of plots on the horizontal axis
    :param verbose: display messages when columns are not visualized
    :param label_color: label color of ticks, titles, x/y axis values / labels.
    """
    # If the column_names argument is not a list then create a list
    if not type(column_names) == list and column_names is not None:
        column_names = [column_names]

    # If we don't have a list of column names then create a histogram for every column.
    if column_names is not None:
        columns = list(dataframe[column_names].columns)
    else:
        columns = dataframe.select_dtypes(include=_np.number).columns.values

    # Calculate the number of rows / columns for the subplot.
    rows = max(int(len(columns) / cols), 1)
    cols = min(cols, len(columns))
    rows += 1 if rows * cols < len(columns) else 0

    # If figsize is not specified then calculate the fig-size
    if figsize is None:
        figsize = (17, rows * 4)

    fig, axs = _plt.subplots(nrows=rows, ncols=cols, figsize=figsize)

    # If we have more then one column then flatten the axis so we can loop through them,
    # if we have only one column then create list containing the axis so we can still loop through it.
    if len(columns) > 1:
        axs = axs.flatten()
    else:
        axs = [axs]

    # Horizontal / vertical padding between the histograms.
    fig.tight_layout(h_pad=10, w_pad=5)
    fig.patch.set_alpha(0.0)  # Transparant background

    n = 0
    for c in columns:
        v = dataframe[c]
        nunique = v.nunique()

        if _pd.api.types.is_numeric_dtype(v) and min_nunique < nunique:
            # Make sure we do not get more then 50 bins.
            nunique = 50 if nunique > 50 else nunique
            bins = 'auto'
            if nunique < 3:
                bins = 2
            _sns.histplot(data=v, kde=bins == 'auto', stat='count', bins=bins, ax=axs[n], kde_kws={'bw_adjust': .15})

            if nunique > 3:
                # Display the mean
                axs[n].text(v.mean(), axs[n].get_ylim()[1] + 10, 'Mean', rotation='vertical', ha='center', va='bottom',
                            color=label_color)

                axs[n].axvline(v.mean(), label=f'Mean ({v.mean():.2f})', color='black', linestyle='dotted')

                # Display the Z-Score
                color = iter(["red", "red", "red", 'red', 'red', 'red'])
                z_score_min = [-3 * v.std() + v.mean(), -2 * v.std() + v.mean(), -1 * v.std() + v.mean()]
                for idx, value in enumerate(z_score_min):
                    # Only show when it is not before the graph starts.
                    c = next(color)
                    if value > v.min():
                        axs[n].text(value, axs[n].get_ylim()[1] + 10, f'-{3 - idx}σ', rotation='vertical', ha='center',
                                    va='bottom', color=label_color)
                        axs[n].axvline(value, label=f'-{3 - idx}σ ({value:.2f})', color=c, linestyle='dotted')

                z_score_max = [1 * v.std() + v.mean(), 2 * v.std() + v.mean(), 3 * v.std() + v.mean()]
                for idx, value in enumerate(z_score_max):
                    # Only show when it is not after the graph starts.
                    c = next(color)
                    if value < v.max():
                        axs[n].text(value, axs[n].get_ylim()[1] + 10, f'{idx + 1}σ', rotation='vertical', ha='center',
                                    va='bottom', color=label_color)
                        axs[n].axvline(value, label=f'{idx + 1}σ ({value:.2f})', color=c, linestyle='dotted')

                # Display the IREQ and lower / upper bounds.
                q1 = v.quantile(q=.25)
                q3 = v.quantile(q=.75)
                irq = (q3 - q1)
                irql = q1 - (1.5 * irq)
                irqu = q3 + (1.5 * irq)

                if axs[n].get_xlim()[1] > irql > axs[n].get_xlim()[0]:
                    axs[n].text(irql, axs[n].get_ylim()[1] + 10, 'IRQ-L', rotation='vertical', ha='center', va='bottom',
                                color=label_color)
                    axs[n].axvline(irql, label=f'IRQ-L ({irql:.2f})', color='green', linestyle='dotted')

                if axs[n].get_xlim()[1] > q1 > axs[n].get_xlim()[0]:
                    axs[n].text(q1, axs[n].get_ylim()[1] + 10, 'Q1', rotation='vertical', ha='center', va='bottom',
                                color=label_color)
                    axs[n].axvline(q1, label=f'Q1 ({q1:.2f})', color='green', linestyle='dotted')

                if axs[n].get_xlim()[1] > q1 > axs[n].get_xlim()[0]:
                    axs[n].text(q3, axs[n].get_ylim()[1] + 10, 'Q3', rotation='vertical', ha='center', va='bottom',
                                color=label_color)
                    axs[n].axvline(q3, label=f'Q3 ({q3:.2f})', color='green', linestyle='dotted')

                if axs[n].get_xlim()[1] > irqu > axs[n].get_xlim()[0]:
                    axs[n].text(irqu, axs[n].get_ylim()[1] + 10, 'IRQ-U', rotation='vertical', ha='center', va='bottom',
                                color=label_color)
                    axs[n].axvline(irqu, label=f'IRQ-U ({irqu:.2f})', color='green', linestyle='dotted')

                axs[n].legend()
            axs[n].grid(False)
        else:
            if verbose:
                print(
                    f"Column '{v.name}' is not visualized, the number of nunique values ({nunique}) either exceeds {max_nunique} or is lower then {min_nunique} or the values are not numeric.")
            continue

        # Only rotate the x labels
        for tick in axs[n].get_xticklabels():
            tick.set_rotation(45)

        # Remove unnecessary white space on the left/right side of the graph.
        axs[n].margins(x=0)
        axs[n].grid(axis='y')

        axs[n].xaxis.label.set_color(label_color)  # Set color of x-axis label
        axs[n].tick_params(axis='x', colors=label_color)  # Set color of x-axis ticks.

        axs[n].yaxis.label.set_color(label_color)  # Set color of y-axis label
        axs[n].tick_params(axis='y', colors=label_color)  # Set color of y-axis ticks.
        axs[n].title.set_color(label_color)  # Set color of title

        n += 1

    for i in range(n, rows * cols):
        fig.delaxes(axs[i])

    _plt.show()


def lineplot(dataframe: _pd.DataFrame, column_names: list[str] = None, figsize: tuple = None, cols=1,
             verbose=1, label_color='black'):
    """
    Plots a histogram for each of the numeric columns in the DataFrame.

    :param dataframe: the pandas dataframe
    :param column_names: columns which exist within the DataFrame if none specified all columns that are numeric
        will be processed
    :param figsize: size of the plot, if None specified then one is calculated
    :param cols: number of plots on the horizontal axis
    :param verbose: display messages when columns are not visualized
    :param label_color: label color of ticks, titles, x/y axis values / labels.
    """
    # assert column_names is not None, "column_names cannot be None"

    # If the column_names argument is not a list then create a list
    if not type(column_names) == list and column_names is not None:
        column_names = [column_names]

    # If we don't have a list of column names then create a histogram for every column.
    if column_names is not None:
        columns = list(dataframe[column_names].columns)
    else:
        columns = dataframe.select_dtypes(include=_np.number).columns.values

    # Calculate the number of rows / columns for the subplot.
    rows = max(int(len(columns) / cols), 1)
    cols = min(cols, len(columns))
    rows += 1 if rows * cols < len(columns) else 0

    # If figsize is not specified then calculate the fig-size
    if figsize is None:
        figsize = (17, rows * 3)

    fig, axs = _plt.subplots(nrows=rows, ncols=cols, figsize=figsize)

    # If we have more then one column then flatten the axis so we can loop through them,
    # if we have only one column then create list containing the axis so we can still loop through it.
    if len(columns) > 1:
        axs = axs.flatten()
    else:
        axs = [axs]

    # Horizontal / vertical padding between the histograms.
    fig.tight_layout(h_pad=10, w_pad=5)
    fig.subplots_adjust(wspace=0, hspace=0.5)
    fig.patch.set_alpha(0.0)  # Transparant background

    n = 0
    for c in columns:
        v = dataframe[c]
        if _pd.api.types.is_numeric_dtype(v):
            _sns.lineplot(data=v, ax=axs[n])
            axs[n].grid(False)
        else:
            if verbose:
                print(
                    f"Column '{v.name}' is not visualized, the values are not numeric.")
            continue

        # Only rotate the x labels
        for tick in axs[n].get_xticklabels():
            tick.set_rotation(45)

        # Remove unnecessary white space on the left/right side of the graph.
        axs[n].margins(x=0)
        axs[n].grid(axis='y')

        axs[n].xaxis.label.set_color(label_color)  # Set color of x-axis label
        axs[n].tick_params(axis='x', colors=label_color)  # Set color of x-axis ticks.

        axs[n].yaxis.label.set_color(label_color)  # Set color of y-axis label
        axs[n].tick_params(axis='y', colors=label_color)  # Set color of y-axis ticks.
        axs[n].title.set_color(label_color)  # Set color of title

        n += 1

    for i in range(n, rows * cols):
        fig.delaxes(axs[i])

    _plt.show()


def correlation(dataframe: _pd.DataFrame, numeric_only=True, method: str = 'pearson', figsize: tuple = (12, 12),
                label_color='black') -> None:
    """
    Create a correlation heatmap for the given DataFrame.

    Parameters:
    - dataframe (pd.DataFrame): The input DataFrame.
    - numeric_only (bool): Whether to include only numeric columns.
    - method (str): The method to use for the correlation heatmap (pearson, spearman or kendall)
    - figsize (tuple): Size of the figure.
    - label_color (str): Color of axis labels and title.

    Returns:
    - None
    """
    # Validate input parameters
    if not isinstance(dataframe, _pd.DataFrame):
        raise ValueError("Input 'dataframe' must be a Pandas DataFrame.")

    if not isinstance(figsize, tuple):
        raise ValueError("'figsize' must be a tuple.")

    if method not in ['pearson', 'kendall', 'spearman']:
        raise ValueError("Invalid correlation method. Supported methods: 'pearson', 'kendall', 'spearman'.")

    # Create the correlation dataset
    df_corr = dataframe.corr(numeric_only=numeric_only, method=method)

    # Create a mask so that only the lower left triangle remains.
    mask = _np.triu(_np.ones_like(df_corr, dtype=bool))

    # Create heatmap.
    fig, ax = _plt.subplots(figsize=figsize)
    fig.patch.set_alpha(0.0)  # Transparent background
    _sns.heatmap(data=df_corr.round(2), mask=mask, vmin=-1, vmax=1, annot=True, cmap='coolwarm',
                 ax=ax)

    # Customize labels and title
    ax.xaxis.label.set_color(label_color)  # Set color of x-axis label
    ax.tick_params(axis='x', colors=label_color)  # Set color of x-axis ticks.

    ax.yaxis.label.set_color(label_color)  # Set color of y-axis label
    ax.tick_params(axis='y', colors=label_color)  # Set color of y-axis ticks.

    ax.title.set_color(label_color)  # Set color of title
    ax.set_title(f'Correlation (method: {method})')
    _plt.show()


def correlation_pb(dataframe: _pd.DataFrame, continuous_columns: list[str], dichotomous_column: str, figsize: tuple = (12, 6),
                   label_color='black') -> None:
    """
    Create a correlation bar plot between continuous features and a dichotomous feature using point-biserial
    correlation coefficient.

    NOTE: A dichotomous feature is a feature which only has two possible values.

    Parameters:
    - dataframe (pd.DataFrame): The input DataFrame with continuous and dichotomous features.
    - continuous_columns (list): List of continuous variable columns.
    - dichotomous_column (str): Column name of the dichotomous variable.
    - figsize (tuple): Size of the figure.
    - label_color (str): Color of axis labels and title.

    Returns:
    - None
    """
    # Validate input parameters
    if not isinstance(dataframe, _pd.DataFrame):
        raise ValueError("Input 'dataframe' must be a Pandas DataFrame.")

    if not isinstance(figsize, tuple):
        raise ValueError("'figsize' must be a tuple.")

    if not isinstance(continuous_columns, list) or not all(isinstance(col, str) for col in continuous_columns):
        raise ValueError("'continuous_columns' must be a list of continuous variable column names.")

    if not isinstance(dichotomous_column, str):
        raise ValueError(
            "'dichotomous_column' must be a string representing the column name of the dichotomous variable.")

    if dichotomous_column not in dataframe.columns:
        raise ValueError(f"'{dichotomous_column}' not found in the DataFrame columns.")

    for col in continuous_columns:
        if col not in dataframe.columns:
            raise ValueError(f"'{col}' not found in the DataFrame columns.")

    # Drop Null columns.
    dataframe = dataframe.dropna()

    # Calculate point-biserial correlation coefficients
    corr_results = [_sp.stats.pointbiserialr(dataframe[col], dataframe[dichotomous_column]) for col in continuous_columns]

    # Extract correlation coefficients and p-values
    corr_coefficients = [result[0] for result in corr_results]
    p_values = [result[1] for result in corr_results]

    # Set up the matplotlib figure
    fig, ax = _plt.subplots(figsize=figsize)
    fig.patch.set_alpha(0.0)  # Transparent background

    # Create a color palette for the bars
    color_palette = _sns.color_palette("coolwarm", len(continuous_columns))

    # Create a bar plot using Seaborn
    _sns.barplot(x=continuous_columns, y=corr_coefficients, ax=ax, palette=color_palette)

    for i, (value, p_value) in enumerate(zip(corr_coefficients, p_values)):
        ax.text(i, value, f"c:{value:.2f}, p:{p_value:.4f}", ha='center', va='bottom')

    # Customize labels and title
    ax.set_xlabel("Continuous Variables", color=label_color)
    ax.xaxis.label.set_color(label_color)
    ax.tick_params(axis='x', colors=label_color)

    ax.set_ylabel("Point-Biserial Correlation Coefficient", color=label_color)
    ax.yaxis.label.set_color(label_color)
    ax.tick_params(axis='y', colors=label_color)

    ax.title.set_color(label_color)
    _plt.title(f'Correlation with {dichotomous_column}')

    _plt.show()



import tensorflow as _tf


def add_batch_to_tensor(x: _tf.Tensor) -> _tf.Tensor:
    """
    Adds a batch size to the given tensor if x = (224, 224, 3) then the result will be (0, 224, 224, 3)

    Args:
        x: The tensor

    Returns:
        A tensor with a batch size of 0.
    """
    return _tf.expand_dims(x, axis=0)


import matplotlib.pyplot as _plt


def xy_data_with_label(x, y) -> None:
    """
    Plots a graph of the values of x whereas x contains vectors of x/y coordinates and y
    is the label (0 or 1).

    Args:
        x: is an array containing vectors of x/y coordinates.
        y: are the associated labels (0=blue, 1=red)
    """
    _plt.plot(x[:, 0][y == 1], x[:, 1][y == 1], "bs")
    _plt.plot(x[:, 0][y == 0], x[:, 1][y == 0], "r^")

    # X contains two features, x1 and x2
    _plt.xlabel(r"$x_1$", fontsize=20)
    _plt.ylabel(r"$x_2$", fontsize=20)

    # Displaying the plot.
    _plt.show()



import numpy as _np
import tensorflow as _tf
import tensorflow.keras as _ks
import so_ml_tools as _soml


def to_categorical(y, num_classes=None, dtype=None):
    """
    Converts a class vector (integers) to binary class matrix (one-hot encoded).

    Args:
        y (array-like): Class vector to be converted into a matrix (integers from 0 to num_classes - 1).
        num_classes (int, optional): Total number of classes. If not provided, the number of classes is inferred from
            the input data. Default is None.
        dtype (str or numpy.dtype, optional): Data type of the output matrix. If not provided, the default data type
            is used. Default is None.

    Returns:
        numpy.ndarray: Binary matrix representation of the input class vector. Each row corresponds to one sample, and
        each column corresponds to one class. The value at (i, j) represents whether sample i belongs to class j,
        where 1 indicates membership and 0 indicates non-membership.

    Raises:
        ValueError: If y contains values outside the range [0, num_classes).
        ValueError: If num_classes is not provided and cannot be inferred from the input data.

    Examples:
        >>> y = [0, 1, 2, 1]
        >>> to_categorical(y)
        array([[1., 0., 0.],
               [0., 1., 0.],
               [0., 0., 1.],
               [0., 1., 0.]], dtype=float32)

    References:
        - Original implementation: https://github.com/keras-team/keras/blob/main/keras/utils/np_utils.py#L11

    """
    return _ks.utils.to_categorical(y, num_classes, dtype=dtype)


def diff_indexes(y_true, y_pred=None, y_prob=None):
    if isinstance(y_true, _tf.data.Dataset):
        y_true = _soml.tf.dataset.get_labels(dataset=y_true)

    y_true = _soml.util.prediction.probability_to_prediction(y=y_true)
    if y_pred is None and y_prob is not None:
        y_pred = _soml.util.prediction.probability_to_prediction(y=y_prob)
    elif y_pred is None and y_prob is None:
        raise "Must specify 'y_pred' or 'y_prob'"

    compare_to = y_pred

    # Check if the shapes match, if not, try to fix it.
    if y_pred.shape != y_true.shape:
        compare_to = _np.reshape(y_pred, newshape=y_true.shape)

    compared_result = y_true == compare_to
    return _np.where(compared_result == False)[0]



import numpy as _np
import so_ml_tools as _soml


def is_one_hot_encoded(value) -> bool:
    """Check if the input array represents a one-hot encoded array.

    Args:
        value (numpy.ndarray or any): Input array to be checked.

    Returns:
        bool: True if the input array represents a one-hot encoded array, False otherwise.

    Raises:
        None

    Examples:
        >>> import numpy as np
        >>> is_one_hot_encoded(np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]]))
        True
        >>> is_one_hot_encoded(np.array([[1, 1, 0], [0, 1, 0], [0, 0, 1]]))
        False
    """
    if not isinstance(value, _np.ndarray):
        value = _soml.util.types.to_numpy(value=value)

    if len(value.shape) != 2:
        return False

    if not _np.all((value == 0) | (value == 1)):
        return False

    if not _np.all(_np.sum(value, axis=1) == 1):
        return False

    return True


def one_hot_to_indices(value) -> _np.ndarray:
    """Converts a one-hot encoded array to a list of indices.

    Args:
        value (numpy.ndarray or any): Input array to be converted.

    Returns:
        numpy.ndarray: Array of indices corresponding to the one-hot encoding.

    Raises:
        ValueError: If the input is not a one-hot encoded array or not a one-dimensional NumPy array.

    Examples:
        >>> import numpy as np
        >>> one_hot_to_indices(np.array([1, 0, 0]))
        [0]
        >>> one_hot_to_indices(np.array([0, 1, 0]))
        [1]
        >>> one_hot_to_indices(np.array([0, 0, 1]))
        [2]
    """
    if not isinstance(value, _np.ndarray):
        value = _soml.util.types.to_numpy(value=value)

    if not _soml.util.onehot.is_one_hot_encoded(value):
        raise ValueError('Input is not recognized as a one-hot encoded value.')

    return _np.argmax(value, axis=1)



import numpy as _np
import pandas as _pd
import tensorflow as _tf
import re as _re
from typing import Union as _Union

########################################################################################################################
# Word level functions
########################################################################################################################

# Based on TextVectorization
PUNCTUATION = r'[!"#$%&()\*\+,-\./:;<=>?@\[\\\]^_`{|}~\']'
LOWER_AND_STRIP_PUNCTUATION = "lower_and_strip_punctuation"
STRIP_PUNCTUATION = "strip_punctuation"
LOWER = "lower"


def count_unique_words(lines: _Union[list[str], _tf.Tensor], standardize="lower_and_strip_punctuation") -> int:
    if _tf.is_tensor(x=lines):
        lines = lines.numpy()

    regex = _re.compile(PUNCTUATION)

    unique = set()
    for line in lines:
        if isinstance(line, (bytes, bytearray)):
            line = line.decode(encoding='utf-8')

        if standardize in (LOWER, LOWER_AND_STRIP_PUNCTUATION):
            line = line.lower()
        if standardize in (STRIP_PUNCTUATION, LOWER_AND_STRIP_PUNCTUATION):
            line = regex.sub('', line)
        unique = unique.union(set(line.split()))
    return len(unique)


def word_count_for_sentences(lines: _Union[list[str], _tf.Tensor]):
    if _tf.is_tensor(x=lines):
        lines = lines.numpy()

    word_count_per_sentence = __count_words_for_each_sentence(lines)
    word_count_for_90_percentile = int(_np.percentile(a=word_count_per_sentence, q=90))
    word_count_for_95_percentile = int(_np.percentile(a=word_count_per_sentence, q=95))
    word_count_for_97_percentile = int(_np.percentile(a=word_count_per_sentence, q=97))
    word_count_for_99_percentile = int(_np.percentile(a=word_count_per_sentence, q=99))
    word_count_for_max = max(word_count_per_sentence)
    word_count_for_average = round(sum(word_count_per_sentence) / len(lines))

    print('Word count for x-percentile sentences:')
    print(f' 90%: {word_count_for_90_percentile}')
    print(f' 95%: {word_count_for_95_percentile}')
    print(f' 97%: {word_count_for_97_percentile}')
    print(f' 99%: {word_count_for_99_percentile}')
    print(f'100%: {word_count_for_max}\n')
    print('Average word count for sentences:')
    print(f'Average: {word_count_for_average}')

########################################################################################################################
# Character level functions
########################################################################################################################

def count_unique_chars(lines: _Union[list[str], _tf.Tensor], standardize="lower_and_strip_punctuation") -> (int, list[chr]):
    regex = _re.compile(PUNCTUATION)

    if _tf.is_tensor(x=lines):
        lines = lines.numpy()

    unique = set()
    for line in lines:
        if isinstance(line, (bytes, bytearray)):
            line = line.decode(encoding='utf-8')

        if standardize in (LOWER, LOWER_AND_STRIP_PUNCTUATION):
            line = line.lower()
        if standardize in (STRIP_PUNCTUATION, LOWER_AND_STRIP_PUNCTUATION):
            line = regex.sub('', line)
        unique = unique.union(set(list(line)))
    return len(unique), unique


def length_per_sentence(lines: _Union[list[str], _tf.Tensor]) -> list[int]:
    """
    Calculates the length for each string value and returns a list containing
    these lengths.

    :param lines: a list of string values
    :return:
    """
    if _tf.is_tensor(x=lines):
        lines = lines.numpy()

    return [len(sentence) for sentence in lines]


def calculate_q_precentile_character_lengths_for_all_sentence(lines: _Union[list[str], _tf.Tensor], q=95) -> int:
    """
    Calculates the q-percentile based on the lengths of the strings.

    :param lines: a list of string values
    :param q: q-percentile (default 95%)
    :return: the percentile
    """
    return int(_np.percentile(a=length_per_sentence(lines), q=q))


def calculate_average_character_length_for_all_sentences(lines: _Union[list[str], _tf.Tensor]) -> int:
    """
    Returns the average length of the strings.

    :param lines: a list of string values
    :return: the average sentence length
    """
    if _tf.is_tensor(x=lines):
        lines = lines.numpy()

    return round(sum(length_per_sentence(lines)) / len(lines))


def __count_words_for_each_sentence(lines: _Union[list[str], _tf.Tensor, _pd.Series]) -> list[int]:
    """
    Calculates the length for each string value and returns a list containing
    these lengths.

    :param lines: a list of string values
    :return:
    """
    if isinstance(lines, _pd.DataFrame):
        raise TypeError("lines is a Pandas DataFrame, select a series to calculate the length")
    if isinstance(lines, _pd.Series):
        lines = lines.numpy()
    if _tf.is_tensor(x=lines):
        lines = lines.numpy()

    return [len(i.split()) for i in lines]



import tensorflow as _tf
import pandas as _pd
import keras as _ks
import numpy as _np
import math as _math
import so_ml_tools as _soml

from typing import Union as _Union


def make_naive_predictions(y_true):
    """
    Method that creates a naive prediction from the given truth labels:

    Input:
        [0, 1, 2, 3, 4, 5, 6, 7]

    Output:
        [1, 2, 3, 4, 5, 6, 7]
        [0, 1, 2, 3, 4, 5, 6]


    Args:
        y_true: the truth labels for which we want to make a naive prediction

    Returns:
        y_true and the naive prediction based on the given truth labels

    """
    naive_predictions = y_true[:-1]
    return y_true[1:], naive_predictions


def dataset_from_array_multivariate(
        data: _Union[list, _np.ndarray, _pd.DataFrame, _tf.Tensor],
        label_column_idx,
        window_size: int,
        horizon_size: int = 1,
        centered: bool = False,
        batch_size=None) -> _tf.data.Dataset:
    """Creates a TensorFlow dataset from a multivariate time series array.

        Args:
            data (list or numpy.ndarray or pandas.Series or pandas.DataFrame or tensorflow.Tensor):
                Input array to be checked.: Input containing the multivariate time series data.
            label_column_idx (int): Index of the column representing the target label.
            window_size (int): Size of the sliding window used for creating sequences.
            horizon_size (int, optional): Size of the prediction horizon. Defaults to 1.
            centered (bool, optional): Whether to center the sliding window around each label.
                Defaults to False.
            batch_size (int, optional): Number of samples per batch. Defaults to None.

        Returns:
            tensorflow.data.Dataset: TensorFlow dataset containing the input sequences and corresponding labels.

        Raises:
            AssertionError: If horizon_size is greater than window_size or if centered is True but window_size is too small.

        Examples:
            >>> import numpy as np
            >>> import tensorflow as tf
            >>> array = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9], [10, 11, 12]])
            >>> dataset = dataset_from_array_multivariate(array, label_column_idx=0, window_size=2, horizon_size=1)
            >>> for batch in dataset:
            ...     print(batch)
            (array([[[1, 2, 3], [4, 5, 6]]]), array([[7], [10]]))
        """
    assert horizon_size < window_size, f"horizon_size '{horizon_size}' must be smaller then window_size '{window_size}'"
    assert not centered or (
            centered and window_size > 2), f"Centered window requested but window_size is too small: '{window_size}', needs to be at least 3 or higher."

    x = _soml.util.types.to_numpy(data)

    label_offset = window_size
    if centered:
        label_offset = _math.floor(window_size / 2)

    y = x[label_offset:, label_column_idx]

    if horizon_size > 1:
        y = [y[start_idx:start_idx + horizon_size] for start_idx, _ in enumerate(y)]

        # Remove last set of lines which are not complete
        y = y[0:-(window_size - horizon_size)]

    return _ks.utils.timeseries_dataset_from_array(
        x,
        y,
        sequence_length=window_size,
        sequence_stride=1,
        batch_size=batch_size)


def dataset_from_array_univariate(
        data: _Union[list, _np.ndarray, _pd.DataFrame, _tf.Tensor],
        window_size: int,
        horizon_size: int = 1,
        centered: bool = False,
        batch_size=None) -> _tf.data.Dataset:
    """Creates a TensorFlow dataset from a univariate time series array.

        Args:
            data (list or numpy.ndarray or pandas.Series or pandas.DataFrame or tensorflow.Tensor): Data array containing the univariate time series data.
            window_size (int): Size of the sliding window used for creating sequences.
            horizon_size (int, optional): Size of the prediction horizon. Defaults to 1.
            centered (bool, optional): Whether to center the sliding window around each label.
                Defaults to False.
            batch_size (int, optional): Number of samples per batch. Defaults to None.

        Returns:
            tensorflow.data.Dataset: TensorFlow dataset containing the input sequences and corresponding labels.

        Raises:
            AssertionError: If horizon_size is greater than window_size or if centered is True but window_size is too small.

        Examples:
            >>> import numpy as np
            >>> import tensorflow as tf
            >>> array = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9])
            >>> dataset = dataset_from_array_univariate(array, window_size=3, horizon_size=1)
            >>> for batch in dataset:
            ...     print(batch)
            (array([[1, 2, 3]]), array([4]))
        """
    assert horizon_size <= window_size, f"horizon_size '{horizon_size}' must be smaller or equal to window_size '{window_size}'"
    assert not centered or (
            centered and window_size > 2), f"Centered window requested but window_size is too small: '{window_size}', needs to be at least 3 or higher."

    x = _soml.util.types.to_numpy(data).ravel()

    label_offset = window_size
    if centered:
        label_offset = _math.floor(window_size / 2)

    y = x[label_offset:]

    if horizon_size > 1:
        y = [y[start_idx:start_idx + horizon_size] for start_idx, _ in enumerate(y)]
        # Remove last set of labels which are not complete horizons
        y = y[:-(window_size - horizon_size)]

    return _ks.utils.timeseries_dataset_from_array(
        x,
        y,
        sequence_length=window_size,
        batch_size=batch_size)



import numpy as np
import tensorflow as _tf
import numpy as _np
from typing import Union as _Union
from typing import Tuple as _Tuple


def split(dataset: _tf.data.Dataset, split_percentages: list[float]) -> tuple[_tf.data.Dataset, ...]:
    """Splits a TensorFlow dataset into multiple datasets based on the provided split percentages.

    Args:
        dataset (tensorflow.data.Dataset): Input TensorFlow dataset to be split.
        split_percentages (List[float]): List of float values representing the percentages of each split.

    Returns:
        Tuple[tensorflow.data.Dataset, ...]: Tuple of TensorFlow datasets after splitting.

    Raises:
        AssertionError: If the sum of split percentages is not equal to 1.0.

    Examples:
        >>> import tensorflow as tf
        >>> dataset = tf.data.Dataset.range(10)
        >>> splits = split(dataset, split_percentages=[0.7, 0.2, 0.1])
        >>> for split in splits:
        ...     print(list(split.as_numpy_iterator()))
        [0, 1, 2, 3, 4, 5, 6], [7, 8], [9]
    """
    assert sum(split_percentages) == 1.0, "The sum of split percentages should be equal to 1.0"

    # Convert percentages to actual lengths from the dataset.
    split_lengths = [int(split_pct * len(dataset)) for split_pct in split_percentages]

    split_datasets = []
    for split_length in split_lengths:
        split_datasets.append(dataset.take(split_length))
        dataset = dataset.skip(split_length)

    return tuple(split_datasets)


def describe_pipeline(dataset: _tf.data.Dataset):
    """
    Describes the different steps in the `tf.data.Dataset` pipeline.

    :param dataset: the dataset.
    """
    input_dataset = dataset
    print(f"{input_dataset.__class__.__name__}", end="")

    while hasattr(input_dataset, '_input_dataset'):
        print(" -> ", end="")
        input_dataset = input_dataset._input_dataset
        print(f"{input_dataset.__class__.__name__}", end="")


def describe_inputs_and_outputs(dataset: _tf.data.Dataset) -> None:
    """
    Describes the input / output shapes and dtype's for the given `tf.data.Dataset`

    Args:
        dataset: a `tf.data.Dataset`
    """
    if isinstance(dataset.element_spec, tuple):
        print(f"{'dtype':<16} | shape")
        for inputs in dataset.element_spec:
            # In case of a single input
            if isinstance(inputs, _tf.TensorSpec):
                print(f"{str(inputs.dtype.name):<16} | {str(inputs.shape)}")
            else:
                # In case of multiple concatenated datasets.
                for i, input in enumerate(inputs):
                    print(f"{str(input.dtype.name):<16} | {str(input.shape)}")


def add_rescaling_mapping(dataset: _tf.data.Dataset) -> _tf.data.Dataset:
    """
    Adds a rescale mapping to the dataset (0-255 -> 0-1)

    Preferably do this before using `optimize_pipeline`

    Args:
        dataset: a `tf.data.Dataset`

    Returns:
        A `tf.data.Dataset` with rescaling.
    """
    return dataset.map(_rescale)


def _rescale(x, y):
    return x / 255., y


def get_class_names_from_dataset_info(ds_info: dict):
    """
    Returns the labels from the dataset info object which is created from loading a
    TensorFlow Dataset (tensorflow_datasets)

    :param ds_info: The dataset info object
    :return: The labels / class names
    """
    if hasattr(ds_info, 'features'):
        return ds_info.features['label'].names

    return ds_info["features"]["label"].names


def get_class_names(dataset: _tf.data.Dataset):
    """
    Returns the class names from the dataset
    :param dataset: the dataset from which we want the class names.
    :return: the class names
    """
    if not isinstance(dataset, _tf.data.Dataset):
        raise TypeError('dataset is not a tf.data.Dataset')

    input_dataset = dataset
    while not hasattr(input_dataset, 'class_names') and hasattr(input_dataset, '_input_dataset'):
        # noinspection PyProtectedMember
        input_dataset = input_dataset._input_dataset

    if not hasattr(input_dataset, 'class_names'):
        raise TypeError("dataset does not have a ´class_names´ attribute defined.")

    return input_dataset.class_names


def get_labels(dataset: _tf.data.Dataset, max_samples: int = None) -> _np.ndarray:
    """
    Returns the labels from a (batched)Dataset

    :param dataset: the dataset from which we want the labels.
    :param max_samples: the maximum number of samples to return
    :return: the labels
    """
    if is_shuffled(dataset=dataset):
        print("WARNING: Dataset is shuffled, retrieving features and labels separate will result in "
              "different order of features and labels")

    if is_batched(dataset=dataset):
        dataset = dataset.unbatch()

    current_sample = 0
    all_labels = None
    itr = dataset.as_numpy_iterator()
    while max_samples is None or current_sample < max_samples:
        try:
            _, labels = next(itr)
            if all_labels is None:
                all_labels = _np.expand_dims(labels, axis=0)
            else:
                all_labels = _np.vstack((all_labels, np.expand_dims(labels, axis=0)))
            current_sample += 1
        except StopIteration:
            break

    return all_labels


def get_features(dataset: _tf.data.Dataset, max_samples: int = None) -> _np.ndarray:
    """
    Returns the features from a (batched)Dataset

    :param dataset: the dataset from which we want the features.
    :param max_samples: the maximum number of samples to return
    :return: the features
    """
    if is_shuffled(dataset=dataset):
        print("WARNING: Dataset is shuffled, retrieving features and labels separate will result in "
              "different order of features and labels")

    if is_batched(dataset=dataset):
        dataset = dataset.unbatch()

    current_sample = 0
    all_features = None
    itr = dataset.as_numpy_iterator()
    while max_samples is None or current_sample < max_samples:
        try:
            features, _ = next(itr)
            if all_features is None:
                all_features = _np.expand_dims(features, axis=0)
            else:
                all_features = _np.vstack((all_features, _np.expand_dims(features, axis=0)))
            current_sample += 1
        except StopIteration:
            break

    return all_features


def get_features_and_labels(dataset: _tf.data.Dataset, max_samples: int = None) -> _Tuple[_np.array, _np.ndarray]:
    """
    Returns the features and labels from a (batched)Dataset as a tupple (features, labels)

    :param dataset: the dataset from which we want the features and labels.
    :param max_samples: the maximum number of samples to return
    :return: the features and labels
    """
    if is_shuffled(dataset=dataset):
        print("WARNING: Dataset is shuffled, retrieved features and labels will match accordingly, however performing"
              " this action twice will result in different order.")

    if is_batched(dataset=dataset):
        dataset = dataset.unbatch()

    current_sample = 0
    all_features = None
    all_labels = None
    itr = dataset.as_numpy_iterator()
    while max_samples is None or current_sample < max_samples:
        try:
            features, labels = next(itr)
            if all_features is None:
                all_features = _np.expand_dims(features, axis=0)
                all_labels = _np.expand_dims(labels, axis=0)
            else:
                all_features = _np.vstack((all_features, _np.expand_dims(features, axis=0)))
                all_labels = _np.vstack((all_labels, _np.expand_dims(labels, axis=0)))
            current_sample += 1
        except StopIteration:
            break

    return all_features, all_labels


def get_batch_dataset(dataset: _tf.data.Dataset) -> _Union[_tf.data.Dataset, None]:
    input_dataset = dataset
    while not input_dataset.__class__.__name__ == '_BatchDataset' and hasattr(input_dataset, '_input_dataset'):
        # noinspection PyProtectedMember
        input_dataset = input_dataset._input_dataset

    if input_dataset.__class__.__name__ == '_BatchDataset':
        return input_dataset

    return None


def get_batch_size(dataset: _tf.data.Dataset) -> int:
    batch_dataset = get_batch_dataset(dataset=dataset)
    if batch_dataset is not None:
        # noinspection PyUnresolvedReferences
        # noinspection PyProtectedMember
        return batch_dataset._batch_size

    return -1


def is_batched(dataset: _tf.data.Dataset) -> bool:
    return get_batch_dataset(dataset=dataset) is not None


def is_prefetched(dataset: _tf.data.Dataset) -> bool:
    if not isinstance(dataset, _tf.data.Dataset):
        raise TypeError('dataset is not a tf.data.Dataset')

    input_dataset = dataset
    while not input_dataset.__class__.__name__ == '_PrefetchDataset' and hasattr(input_dataset, '_input_dataset'):
        # noinspection PyProtectedMember
        input_dataset = input_dataset._input_dataset

    return input_dataset.__class__.__name__ == '_PrefetchDataset'


def is_cached(dataset: _tf.data.Dataset) -> bool:
    if not isinstance(dataset, _tf.data.Dataset):
        raise TypeError('dataset is not a tf.data.Dataset')

    input_dataset = dataset
    while not input_dataset.__class__.__name__ == 'CacheDataset' and hasattr(input_dataset, '_input_dataset'):
        # noinspection PyProtectedMember
        input_dataset = input_dataset._input_dataset

    return input_dataset.__class__.__name__ == 'CacheDataset'


def is_shuffled(dataset: _tf.data.Dataset) -> bool:
    if not isinstance(dataset, _tf.data.Dataset):
        raise TypeError('dataset is not a tf.data.Dataset')

    input_dataset = dataset
    while not input_dataset.__class__.__name__ == '_ShuffleDataset' and hasattr(input_dataset, '_input_dataset'):
        # noinspection PyProtectedMember
        input_dataset = input_dataset._input_dataset

    return input_dataset.__class__.__name__ == '_ShuffleDataset'



import matplotlib.pyplot as _plt
import seaborn as _sns
import so_ml_tools as _soml


def plot_predictions(x, y_true, y_prob, start_index=None, end_index=None, title='Predictions', figsize=(30, 8)):
    fig, ax = _plt.subplots(figsize=figsize)

    x = _soml.util.types.to_numpy(x)
    y_true = _soml.util.types.to_numpy(y_true)
    y_prob = _soml.util.types.to_numpy(y_prob)

    if start_index is None:
        start_index = 0

    if end_index is None:
        end_index = len(x)

    assert end_index > start_index, 'end_index must be larger then start_index'

    print(f'{start_index}, {end_index}')

    _sns.lineplot(x=x.ravel()[start_index:end_index],
                  y=y_true.ravel()[start_index:end_index],
                  ax=ax, label='Actual')

    _sns.lineplot(x=x.ravel()[start_index:end_index],
                  y=y_prob.ravel()[start_index:end_index],
                  ax=ax, label='Predicted')

    _plt.title(f'{title} (range: {start_index}-{end_index})')
    _plt.legend()
    ax.set_ylabel("Value")
    ax.set_xlabel("Time")
    _plt.show()



import numpy as _np
import tensorflow as _tf
import pandas as _pd
import so_ml_tools as _soml

from typing import Union as _Union


def show_wrong_predicted_images(
        x: _Union[list[list], _np.ndarray, _tf.Tensor],
        y_true: _Union[list[list], _np.ndarray, _tf.Tensor, _pd.Series],
        y_pred: _Union[list[list], _np.ndarray, _tf.Tensor, _pd.Series] = None,
        y_prob: _Union[list[list], _np.ndarray, _tf.Tensor] = None,
        class_names: list[str] = None,
        shape=(4, 8)):
    if _soml.util.onehot.is_one_hot_encoded(value=y_true):
        y_true = _soml.util.onehot.one_hot_to_indices(value=y_true)
        y_true = _np.reshape(a=y_true, newshape=(len(y_true), 1))

    y_true = _soml.util.prediction.probability_to_prediction(y=y_true)
    if y_pred is None and y_prob is not None:
        y_pred = _soml.util.prediction.probability_to_prediction(y=y_prob, maintain_shape=True)
    elif y_pred is None and y_prob is None:
        raise "Must specify 'y_pred' or 'y_prob'"

    x = _soml.util.types.to_numpy(value=x)

    # Get the indices of the labels that are different
    indices = _soml.util.label.diff_indexes(y_true=y_true, y_pred=y_pred)
    X_test_failed = x[indices]
    y_test_failed = y_pred[indices]
    _soml.data.image.show_images_from_nparray_or_tensor(x=X_test_failed, y=y_test_failed, class_names=class_names, shape=shape)



from . import (
    data,
    nlp,
    pd,
    tf,
    util,
    evaluate,
    shap,
    sklearn,
    feature_selection,
    imblearn,
    timeseries
)

# We follow Semantic Versioning (https://semver.org/spec/v2.0.0.html)
_MAJOR_VERSION = '0'
_MINOR_VERSION = '1'
_PATCH_VERSION = '197'

__version__ = '.'.join([
    _MAJOR_VERSION,
    _MINOR_VERSION,
    _PATCH_VERSION,
])

__all__ = [
    "data",
    "nlp",
    "pd",
    "shap",
    "tf",
    "util",
    "evaluate",
    "sklearn",
    "feature_selection",
    "imblearn",
    "timeseries",
    "__version__"
]


import numpy as _np
import pandas as _pd
import tensorflow as _tf
from typing import Union as _Union


def to_numpy(value: _Union[list, _np.ndarray, _pd.Series, _pd.DataFrame, _tf.Tensor]) -> _np.ndarray:
    """Converts input data to a NumPy array.

        Args:
            value (list or numpy.ndarray or pandas.Series or pandas.DataFrame or tensorflow.Tensor):
                Input data to be converted to a NumPy array.

        Returns:
            numpy.ndarray: Converted NumPy array.

        Raises:
            ValueError: If the input data type is unsupported.
    """
    if isinstance(value, list):
        return _np.array(value)
    elif isinstance(value, _np.ndarray):
        return value
    elif isinstance(value, (_pd.DataFrame, _pd.Series)):
        return value.to_numpy()
    elif isinstance(value, _tf.Tensor):
        return value.numpy()
    elif hasattr(value, 'numpy'):
        return value.numpy()
    elif hasattr(value, 'to_numpy'):
        return value.to_numpy()

    raise ValueError("Unsupported data type. Please provide a tensor, dataset, dataframe, or series.")


import shap as _sh
import numpy as _np
import tensorflow as _tf
from typing import Union as _Union


def create_kernel_explainer(model: _Union[_tf.keras.Model, _tf.keras.Sequential], x: _np.array, sample_size=None) \
        -> _sh.Explainer:
    """
    Args:
        model: keras model
        x: training dataset, either a numpy array or a pandas dataframe.
        sample_size:  sample size

    Returns:
        The explainer object.
    """
    x_sample = x
    if sample_size is not None:
        x_sample = x[:sample_size, :]

    return _sh.KernelExplainer(model, x_sample)


def calculate_shap_values(explainer: _sh.Explainer, x: _np.array, n_samples: _Union[int, str] = "auto"):
    """
    Args:
        explainer:
        x:
        n_samples:
    """
    print(f"Calculating shap values for {len(x)} entries.")
    return explainer.shap_values(x, nsamples=n_samples)


def waterfall_plot(explainer, shap_values: list[_np.array], feature_names: list[str], max_features: _Union[str, int] = 'max',
                   display_entry: int = 0, display_class: int = 0):
    """
    Plot the breakdown of the

    Args:
      explainer: the explainer object
      shap_values: calculated values, use shap_values[x] for single class.
      feature_names: a `list` containing the feature names.
      max_features: default 'auto' to show the default amount, set to 'max' to show all
      display_entry: (optional) the entry to display when multiple have been calculated
      display_class: (optional) the class to plot

    Returns:
        None
    """
    if len(shap_values) > 1:  # Multi-class
        shap_value_class = shap_values[display_class]
    else:  # Single-class
        shap_value_class = shap_values[0]

    shap_example_within_class = shap_value_class
    if shap_value_class.ndim > 1:
        shap_example_within_class = shap_value_class[display_entry]

    explanation = _sh.Explanation(
        values=shap_example_within_class,  # Actual values of specific class
        base_values=explainer.expected_value[display_class],  # Mean value: E[f(x)] for specific class.
        feature_names=feature_names)

    if max_features == 'max':
        max_features = len(feature_names)

    _sh.waterfall_plot(shap_values=explanation, max_display=max_features)


def summary_plot(shap_values: list[_np.array], class_names: list[str], feature_names: list[str],
                 plot_type: str = 'dot', display_entry: int = None, display_class: int = None) -> None:
    """
    Summary plot for displaying feature for all entries.

    Args:
      shap_values: calculated values, use shap_values[x] for single class.
      class_names: a `list` containing the class names.
      feature_names: a `list` containing the feature names.
      plot_type: 'bar' for multiclass and single class, 'dot', 'violin' and  for single class.

    Returns:
        None
    """
    assert len(class_names) == len(
        shap_values), (f"Number of class_names ({len(class_names)}) does not match the number of "
                       f"classes in shap_values ({shap_values[0].ndim})")
    assert display_class is None or (
                -1 < display_class < len(class_names)), f"Invalid display_class value {display_class}."

    shap_values = _np.array(shap_values)

    # In case the shap_values were calculated for a single set of features we need to add another dimension.
    if shap_values[0].ndim == 1:
        shap_values = _np.array(list(map(lambda x: _np.expand_dims(x, axis=0), shap_values)))

    # Filter on a specific entry when applicable.
    if display_entry is not None:
        shap_values = _np.expand_dims(shap_values[:, display_entry], axis=1)

    # Filter on a specific class when applicable.
    if display_class is not None:
        class_names = [class_names[display_class]]
        shap_values = _np.expand_dims(shap_values[display_class, :], axis=0)

    # Convert the numpy array back to a list with arrays as was the original data structure
    shap_values = list(map(lambda x: x, shap_values))

    # Old code that was used when dealing with single / multiple entries
    # if shap_values[0].ndim == 1:
    #     shap_values = list(map(lambda x: np.expand_dims(x, axis=0), shap_values))

    # Do we display a single class
    if len(shap_values) == 1:
        if plot_type is None:
            print("Using plot_type = 'dot', other options are 'violin' and 'bar'")
            plot_type = 'dot'

        shap_values = shap_values[0]
    elif plot_type != 'bar':
        print(f"{'Overriding' if plot_type != 'bar' else 'Using'} plot_type with 'bar' since we have multiple classes.")
        plot_type = 'bar'

    _sh.summary_plot(shap_values=shap_values,
                     class_names=class_names,
                     feature_names=feature_names,
                     plot_type=plot_type, show=True)


def force_plot(shap_values: list[_np.array], explainer: _sh.Explainer, feature_names: list[str],
               display_class: int = 0) -> object:
    """
    Force plot for displaying feature for a single class.

    Args:
      shap_values: calculated values
      explainer: the explainer object
      feature_names: a `list` containing the feature names.
      display_class: (optional) the class to plot

    Returns:
        None
    """
    if display_class is None:
        display_class = 0

    return _sh.force_plot(base_value=explainer.expected_value[display_class], shap_values=shap_values[display_class],
                   features=feature_names)
